/*
 * Copyright 2009 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of either or both of the following licenses:
 *
 * 1) the GNU Lesser General Public License version 3, as published by the
 * Free Software Foundation; and/or
 * 2) the GNU Lesser General Public License version 2.1, as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranties of
 * MERCHANTABILITY, SATISFACTORY QUALITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the applicable version of the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of both the GNU Lesser General Public
 * License version 3 and version 2.1 along with this program.  If not, see
 * <http://www.gnu.org/licenses/>
 *
 * Authored by: Jay Taoko <jay.taoko@canonical.com>
 *
 */
/** 
 * SECTION:ctk-utils
 * @short_description: Utility functions
 *
 * Utility functions
 */

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <math.h>

#ifdef WITH_GLES
  // Hearders for OpenGL 2.0: gl2.h, gl2ext.h
  #include <GLES2/gl2.h>
  #include <GLES2/gl2ext.h>
#else
  #include <GL/glew.h>
  #include <GL/glxew.h>
#endif

#include <clutter/clutter.h>
#include "ctk-utils.h"
#include "ctk-gfx-private.h"
#include <gio/gio.h>

/* Whether to display information about coordinate normalization */
#define DEBUG_COORD_NORMALIZATION 0

void print_shader_object_error(GLuint obj)
{
  int info_length = 0;
  int chars_written  = 0;
  char *infoLog;

  if (glIsShader(obj) == GL_TRUE)
    glGetShaderiv(obj, GL_INFO_LOG_LENGTH, &info_length);
  else
    glGetProgramiv(obj, GL_INFO_LOG_LENGTH, &info_length);

  if (info_length > 1)
    {
      infoLog = (char *)malloc(info_length);
      if (glIsShader(obj) == GL_TRUE)
        {
          glGetShaderInfoLog(obj, info_length, &chars_written, infoLog);
          g_warning ("Shader Error Log: %s\n", infoLog);
        }
      else
        {
          glGetProgramInfoLog(obj, info_length, &chars_written, infoLog);
          g_warning ("Program Error Log: %s\n", infoLog);
        }
      free(infoLog);
    }
}

#ifdef WITH_GLES

char *ctk_read_shader_text_file(const char *filename)
{
  GError *error = 0;
  gchar *buffer = 0;
  gsize length = 0;
  g_file_get_contents(filename, &buffer, &length, &error);
  if (error != NULL)
    {
      g_warning("Unable to read file %s: %s\n", filename, error->message);
      g_error_free (error);
      return NULL;
    }
  return buffer;
}

ShaderProgram *ctk_create_shader_program_from_source(const char* vtx_filename, const char* frag_filename)
{
  char *vstxt, *fstxt;
  ShaderProgram* sh = (ShaderProgram*) g_malloc0(sizeof(ShaderProgram));

  sh->shvert = glCreateShader(GL_VERTEX_SHADER);
  vstxt = ctk_read_shader_text_file(vtx_filename);
  glShaderSource(sh->shvert, 1, (const char **)&vstxt, NULL);
  g_free(vstxt);
  glCompileShader(sh->shvert);
  print_shader_object_error(sh->shvert);

  sh->shfrag = glCreateShader(GL_FRAGMENT_SHADER);
  fstxt = ctk_read_shader_text_file (frag_filename);
  glShaderSource(sh->shfrag, 1, (const char **)&fstxt, NULL);
  g_free(fstxt);
  glCompileShader(sh->shfrag);
  print_shader_object_error(sh->shfrag);

  sh->shprog = glCreateProgram();
  glAttachShader(sh->shprog, sh->shvert);
  glAttachShader(sh->shprog, sh->shfrag);
  glLinkProgram(sh->shprog);

  return sh;
}

void ctk_delete_shader_program(ShaderProgram *sh)
{
  if (sh)
    {
      glDetachShader(sh->shprog, sh->shvert);
      glDetachShader(sh->shprog, sh->shfrag);
      glDeleteShader(sh->shvert);
      glDeleteShader(sh->shfrag);
      glDeleteProgram(sh->shprog);
      g_free(sh);
    }
}

void ctk_render_quad_rt(CtkRenderTarget* rt, ShaderProgram *shader, int window_w, int window_h, int x, int y, int w, int h,
	gfloat red, gfloat green, gfloat blue, gfloat Opacity)
{
  guint texid = ctk_render_target_get_color_buffer_ogl_id(rt);
  guint texwidth = ctk_render_target_get_width(rt);
  guint texheight = ctk_render_target_get_height(rt);
  ctk_render_quad(texid, texwidth, texheight, shader, window_w, window_h, x, y, w, h,  red, green, blue, Opacity);
}

void ctk_render_quad(guint texid, guint texwidth, guint texheight, ShaderProgram *shader, int window_w, int window_h, int x, int y, int w, int h,
	gfloat red, gfloat green, gfloat blue, gfloat Opacity)
{
  ClutterVertex vtx[4];

  vtx[0].x = x;
  vtx[0].y = y;
  vtx[1].x = x;
  vtx[1].y = y+h;
  vtx[2].x = x+w;
  vtx[2].y = y+h;
  vtx[3].x = x+w;
  vtx[3].y = y;

  ctk_render_custom_quad(texid, texwidth, texheight, shader, window_w, window_h, vtx, red, green, blue, Opacity);
}

void ctk_render_custom_quad(guint texid, guint texwidth, guint texheight, ShaderProgram *shader, int window_w, int window_h, ClutterVertex vtx[4],
	gfloat red, gfloat green, gfloat blue, gfloat Opacity)
{
  GLint loc;
  GLint posId, texCoordId;

  GLfloat texture_coords[] =
  {
    0.0f, 1.0f,
    0.0f, 0.0f,
    1.0f, 0.0f,
    1.0f, 1.0f,
  };

  GLfloat vertex_positions[] =
  {
    (2.0*vtx[0].x-window_w)/window_w, -(2.0*vtx[0].y-window_h)/window_h, 0.0f, 1.0f,
    (2.0*vtx[1].x-window_w)/window_w, -(2.0*vtx[1].y-window_h)/window_h, 0.0f, 1.0f,
    (2.0*vtx[2].x-window_w)/window_w, -(2.0*vtx[2].y-window_h)/window_h, 0.0f, 1.0f,
    (2.0*vtx[3].x-window_w)/window_w, -(2.0*vtx[3].y-window_h)/window_h, 0.0f, 1.0f,
  };

#if DEBUG_COORD_NORMALIZATION == 1
  g_message("CustomQuad: (%f, %f) (%f, %f) (%f, %f) (%f, %f) =>\n"
            "            (%f, %f) (%f, %f) (%f, %f) (%f, %f)\n"
            "            window_w: %d window_h: %d",
            vtx[0].x, vtx[0].y,
            vtx[1].x, vtx[1].y,
            vtx[2].x, vtx[2].y,
            vtx[3].x, vtx[3].y,
            vertex_positions[0], vertex_positions[1],
            vertex_positions[4], vertex_positions[5],
            vertex_positions[8], vertex_positions[9],
            vertex_positions[12], vertex_positions[13],
            window_w, window_h);
#endif


  /* Set texture 0 environment mode */
  {
    CHECKGL( glActiveTexture(GL_TEXTURE0) );
    CHECKGL( glBindTexture(GL_TEXTURE_2D, texid) );
  }

  /* Use the shader */
  CHECKGL( glUseProgram(shader->shprog) );
  loc = glGetUniformLocation(shader->shprog, "Tex0");
  CHECKGL_MSG( "glGetUniformLocation" );
  CHECKGL( glUniform1i(loc, 0) );
  loc = glGetUniformLocation(shader->shprog, "TextureSize");
  CHECKGL_MSG( "glGetUniformLocation" );
  CHECKGL( glUniform2f(loc, texwidth, texheight) );
  loc = glGetUniformLocation(shader->shprog, "color");
  CHECKGL( glUniform4f(loc, Opacity*red, Opacity*green, Opacity*blue, Opacity) );

  posId = glGetAttribLocation(shader->shprog, "a_position");
  texCoordId = glGetAttribLocation(shader->shprog, "a_texcoord0");

  CHECKGL( glVertexAttribPointer(posId, 4, GL_FLOAT, 0, 0, vertex_positions) );
  CHECKGL( glVertexAttribPointer(texCoordId, 2, GL_FLOAT, 0, 0, texture_coords) );

  CHECKGL( glEnableVertexAttribArray(posId) );
  CHECKGL( glEnableVertexAttribArray(texCoordId) );

  CHECKGL( glDrawArrays(GL_TRIANGLE_FAN, 0, 4) );

  CHECKGL( glDisableVertexAttribArray(posId) );
  CHECKGL( glDisableVertexAttribArray(texCoordId) );

  /* Stop using the shader program */
  CHECKGL( glUseProgram(0) );

  /* Unbind texture */
  CHECKGL( glActiveTexture(GL_TEXTURE0) );
  CHECKGL( glBindTexture(GL_TEXTURE_2D, 0) );

}

void ctk_render_quad_alpha_mask(
      unsigned int texid,
      unsigned int texture_width,
      unsigned int texture_height,
      gfloat red, gfloat green, gfloat blue, gfloat alpha,
      int window_w, int window_h, int x, int y, int w, int h)
{
  GLint loc;
  GLint posId, texCoordId;

  GLfloat texture_coords[] =
  {
    0.0f, 1.0f,
    1.0f, 1.0f,
    1.0f, 0.0f,
    0.0f, 0.0f,
  };

  GLfloat vertex_positions[] =
  {
    (2.0*x-window_w)/window_w, -(2.0*y-window_h)/window_h, 0.0f, 1.0f,
    (2.0*(x+w)-window_w)/window_w, -(2.0*y-window_h)/window_h, 0.0f, 1.0f,
    (2.0*(x+w)-window_w)/window_w, -(2.0*(y+h)-window_h)/window_h, 0.0f, 1.0f,
    (2.0*x-window_w)/window_w, -(2.0*(y+h)-window_h)/window_h, 0.0f, 1.0f,
  };

#if DEBUG_COORD_NORMALIZATION == 1
  g_message("AMask: (%d, %d) (%d, %d) (%d, %d) (%d, %d) =>\n"
            "       (%f, %f) (%f, %f) (%f, %f) (%f, %f)\n"
            "       window_w: %d window_h: %d",
            x, y,
            x + w, y,
            x + w, y + h,
            x , y + h,
            vertex_positions[0], vertex_positions[1],
            vertex_positions[4], vertex_positions[5],
            vertex_positions[8], vertex_positions[9],
            vertex_positions[12], vertex_positions[13],
            window_w, window_h);
#endif

  /* Set texture 0 environment mode */
  {
    CHECKGL( glActiveTexture(GL_TEXTURE0) );
    CHECKGL( glBindTexture(GL_TEXTURE_2D, texid));
  }

  /* Use the shader */
  CHECKGL( glUseProgram(g_shTextureAlpha->shprog) );
  loc = glGetUniformLocation(g_shTextureAlpha->shprog, "Tex0");
  CHECKGL_MSG( "glGetUniformLocation" );
  CHECKGL( glUniform1i(loc, 0) );
  loc = glGetUniformLocation(g_shTextureAlpha->shprog, "TextureSize");
  CHECKGL_MSG( "glGetUniformLocation" );
  CHECKGL( glUniform2f(loc, texture_width, texture_height) );
  loc = glGetUniformLocation(g_shTextureAlpha->shprog, "MaskColor");
  CHECKGL_MSG( "glGetUniformLocation" );
  CHECKGL( glUniform4f(loc, 
            (gfloat)red,
            (gfloat)green,
            (gfloat)blue,
            (gfloat)alpha) );

  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  posId = glGetAttribLocation(g_shTextureAlpha->shprog, "a_position");
  texCoordId = glGetAttribLocation(g_shTextureAlpha->shprog, "a_texcoord0");

  CHECKGL( glVertexAttribPointer(posId, 4, GL_FLOAT, 0, 0, vertex_positions) );
  CHECKGL( glVertexAttribPointer(texCoordId, 2, GL_FLOAT, 0, 0, texture_coords) );

  CHECKGL( glEnableVertexAttribArray(posId) );
  CHECKGL( glEnableVertexAttribArray(texCoordId) );

  CHECKGL( glDrawArrays(GL_TRIANGLE_FAN, 0, 4) );

  CHECKGL( glDisableVertexAttribArray(posId) );
  CHECKGL( glDisableVertexAttribArray(texCoordId) );

  /* Stop using the shader program */
  CHECKGL( glUseProgram(0) );

  /* Unbind texture */
  CHECKGL( glActiveTexture(GL_TEXTURE0) );
  CHECKGL( glBindTexture(GL_TEXTURE_2D, 0) );
}

void custom_render_quad_gaussian_blur_separable(CtkRenderTarget* rt,
                                    ShaderProgram *shader, gfloat sigma,
                                    int window_w, int window_h, int x, int y, int w, int h)
{
  GLint loc;
  GLint posId, texCoordId;

  GLfloat texture_coords[] =
  {
    0.0f, 1.0f,
    1.0f, 1.0f,
    1.0f, 0.0f,
    0.0f, 0.0f,
  };

  GLfloat vertex_positions[] =
  {
    (2.0*x-window_w)/window_w, -(2.0*y-window_h)/window_h, 0.0f, 1.0f,
    (2.0*(x+w)-window_w)/window_w, -(2.0*y-window_h)/window_h, 0.0f, 1.0f,
    (2.0*(x+w)-window_w)/window_w, -(2.0*(y+h)-window_h)/window_h, 0.0f, 1.0f,
    (2.0*x-window_w)/window_w, -(2.0*(y+h)-window_h)/window_h, 0.0f, 1.0f,
  };

  guint num_tap = 7; /* must be odd number */
  gfloat W[num_tap];
  gfloat sum = 0.0f;
  guint i = 0;

  guint half = (num_tap-1)/2;

  W[half] = (1.0f/(sqrt(2.0f*3.14159265358f)*sigma)) * exp(-0.0f/(2.0f*sigma*sigma));
  sum += W[half];
  for(i = 0; i < half; i++)
  {
    float X = (i + 1)*(i + 1);
    W[half - i - 1] = W[half + i + 1] = (1.0f/(sqrt(2.0f*3.14159265358f)*sigma)) * exp(-X/(2.0f*sigma*sigma));
    sum += 2.0f * W[half - i - 1];
  }

  /* normalization */
  for(i = 0; i < num_tap; i++)
  {
    W[i] = W[i] / sum;
  }

  /* Set texture 0 environment mode */
  {
    CHECKGL( glActiveTexture(GL_TEXTURE0) );
    CHECKGL( glBindTexture(GL_TEXTURE_2D, ctk_render_target_get_color_buffer_ogl_id(rt)) );
  }

  /* Use the shader */
  CHECKGL( glUseProgram(shader->shprog) );
  loc = glGetUniformLocation(shader->shprog, "Tex0");
  CHECKGL_MSG( "glGetUniformLocation" );
  CHECKGL( glUniform1i(loc, 0) );
  loc = glGetUniformLocation(shader->shprog, "TextureSize");
  CHECKGL_MSG( "glGetUniformLocation" );
  CHECKGL( glUniform2f(loc, ctk_render_target_get_width(rt), ctk_render_target_get_height(rt)) );
  loc = glGetUniformLocation(shader->shprog, "W");
  CHECKGL_MSG( "glGetUniformLocation" );
  CHECKGL( glUniform1fv(loc, num_tap, W) );
  loc = glGetUniformLocation(shader->shprog, "color");
  CHECKGL( glUniform4f(loc, 1.0f, 1.0f, 1.0f, 1.0f) );

  posId = glGetAttribLocation(g_shTextureAlpha->shprog, "a_position");
  texCoordId = glGetAttribLocation(g_shTextureAlpha->shprog, "a_texcoord0");

  CHECKGL( glVertexAttribPointer(posId, 4, GL_FLOAT, 0, 0, vertex_positions) );
  CHECKGL( glVertexAttribPointer(texCoordId, 2, GL_FLOAT, 0, 0, texture_coords) );

  CHECKGL( glEnableVertexAttribArray(posId) );
  CHECKGL( glEnableVertexAttribArray(texCoordId) );

  CHECKGL( glDrawArrays(GL_TRIANGLE_FAN, 0, 4) );

  CHECKGL( glDisableVertexAttribArray(posId) );
  CHECKGL( glDisableVertexAttribArray(texCoordId) );

  /* Stop using the shader program */
  CHECKGL( glUseProgram(0) );

  /* Unbind texture */
  CHECKGL( glActiveTexture(GL_TEXTURE0) );
  CHECKGL( glBindTexture(GL_TEXTURE_2D, 0) );
}

void custom_render_quad_texture_mask(unsigned int base_texid,
      unsigned int texid,
      unsigned int texture_width,
      unsigned int texture_height,
      ShaderProgram *shader,
      int window_w,
      int window_h,
      int x,
      int y,
      int w,
      int h,
      float Opacity)
{
  GLint loc;
  GLint posId, texCoord0Id, texCoord1Id;

  float tile_u = (float)w/texture_width;
  float tile_v = (float)h/texture_height;

  GLfloat texture_coords0[] =
  {
    0.0f, 1.0f,
    1.0f, 1.0f,
    1.0f, 0.0f,
    0.0f, 0.0f,
  };

  GLfloat texture_coords1[] =
  {
    0.0f, tile_v,
    tile_u, tile_v,
    tile_u, 0.0f,
    0.0f, 0.0f,
  };

  GLfloat vertex_positions[] =
  {
    (2.0*x-window_w)/window_w, -(2.0*y-window_h)/window_h, 0.0f, 1.0f,
    (2.0*(x+w)-window_w)/window_w, -(2.0*y-window_h)/window_h, 0.0f, 1.0f,
    (2.0*(x+w)-window_w)/window_w, -(2.0*(y+h)-window_h)/window_h, 0.0f, 1.0f,
    (2.0*x-window_w)/window_w, -(2.0*(y+h)-window_h)/window_h, 0.0f, 1.0f,
  };

#if DEBUG_COORD_NORMALIZATION == 1
  g_message("TMask: (%d, %d) (%d, %d) (%d, %d) (%d, %d) =>\n"
            "       (%f, %f) (%f, %f) (%f, %f) (%f, %f)\n"
            "       window_w: %d window_h: %d",
            x, y,
            x + w, y,
            x + w, y + h,
            x , y + h,
            vertex_positions[0], vertex_positions[1],
            vertex_positions[4], vertex_positions[5],
            vertex_positions[8], vertex_positions[9],
            vertex_positions[12], vertex_positions[13],
            window_w, window_h);
#endif

  /* Use the shader */
  CHECKGL( glUseProgram(shader->shprog) );

  /* Set texture 0 environment mode */
  {
    CHECKGL( glActiveTexture(GL_TEXTURE0) );
    CHECKGL( glBindTexture(GL_TEXTURE_2D, base_texid) );
  }

  /*  Set texture 1 environment mode */
  {
    CHECKGL( glActiveTexture(GL_TEXTURE1) );
    CHECKGL( glBindTexture(GL_TEXTURE_2D, texid) );
  }

  loc = glGetUniformLocation(shader->shprog, "Tex0");
  CHECKGL_MSG( "glGetUniformLocation" );
  CHECKGL( glUniform1i(loc, 0) );

  loc = glGetUniformLocation(shader->shprog, "Tex1");
  CHECKGL_MSG( "glGetUniformLocation" );
  CHECKGL( glUniform1i(loc, 1) );

  loc = glGetUniformLocation(shader->shprog, "TextureSize");
  CHECKGL_MSG( "glGetUniformLocation" );
  CHECKGL( glUniform2f(loc, texture_width, texture_height) );

  loc = glGetUniformLocation(shader->shprog, "color");
  CHECKGL( glUniform4f(loc, Opacity, Opacity, Opacity, Opacity) );

  posId = glGetAttribLocation(shader->shprog, "a_position");
  texCoord0Id = glGetAttribLocation(shader->shprog, "a_texcoord0");
  texCoord1Id = glGetAttribLocation(shader->shprog, "a_texcoord1");

  CHECKGL( glVertexAttribPointer(posId, 4, GL_FLOAT, 0, 0, vertex_positions) );
  CHECKGL( glVertexAttribPointer(texCoord0Id, 2, GL_FLOAT, 0, 0, texture_coords0) );
  CHECKGL( glVertexAttribPointer(texCoord1Id, 2, GL_FLOAT, 0, 0, texture_coords1) );

  CHECKGL( glEnableVertexAttribArray(posId) );
  CHECKGL( glEnableVertexAttribArray(texCoord0Id) );
  CHECKGL( glEnableVertexAttribArray(texCoord1Id) );

  CHECKGL( glDrawArrays(GL_TRIANGLE_FAN, 0, 4) );

  CHECKGL( glDisableVertexAttribArray(posId) );
  CHECKGL( glDisableVertexAttribArray(texCoord0Id) );
  CHECKGL( glDisableVertexAttribArray(texCoord1Id) );

  /* Stop using the shader program */
  CHECKGL( glUseProgram(0) );

  /* Unbind texture */
  CHECKGL( glActiveTexture(GL_TEXTURE0) );
  CHECKGL( glBindTexture(GL_TEXTURE_2D, 0) );
  CHECKGL( glActiveTexture(GL_TEXTURE1) );
  CHECKGL( glBindTexture(GL_TEXTURE_2D, 0) );
}

void ctk_copy_rendertarget_to_rendertarget(CtkRenderTarget* rt_src, float u0, float v0, float u1, float v1,
                    CtkRenderTarget* rt_dst, int x_dst, int y_dst, int w_dst, int h_dst)
{
  GLint loc;
  GLint posId, texCoordId;
  int window_w = ctk_render_target_get_width(rt_dst);
  int window_h = ctk_render_target_get_height(rt_dst);

  GLfloat texture_coords[] =
  {
    u0, v0,
    u0, v1,
    u1, v1,
    u1, v0,
  };

  GLfloat vertex_positions[] =
  {
    (2.0*x_dst-window_w)/window_w, -(2.0*y_dst-window_h)/window_h, 0.0f, 1.0f,
    (2.0*x_dst-window_w)/window_w, -(2.0*(y_dst+h_dst)-window_h)/window_h, 0.0f, 1.0f,
    (2.0*(x_dst+w_dst)-window_w)/window_w, -(2.0*(y_dst+h_dst)-window_h)/window_h, 0.0f, 1.0f,
    (2.0*(x_dst+w_dst)-window_w)/window_w, -(2.0*y_dst-window_h)/window_h, 0.0f, 1.0f,
  };

#if DEBUG_COORD_NORMALIZATION == 1
  g_message("RT2RT: (%d, %d) (%d, %d) (%d, %d) (%d, %d) =>\n"
            "       (%f, %f) (%f, %f) (%f, %f) (%f, %f)\n"
            "       window_w: %d window_h: %d",
            x_dst, y_dst,
            x_dst, y_dst + h_dst,
            x_dst + w_dst, y_dst + h_dst,
            x_dst + w_dst, y_dst,
            vertex_positions[0], vertex_positions[1],
            vertex_positions[4], vertex_positions[5],
            vertex_positions[8], vertex_positions[9],
            vertex_positions[12], vertex_positions[13],
            window_w, window_h);
#endif

  ctk_render_target_bind(rt_dst);
  /* Use the shader */
  CHECKGL( glUseProgram(g_shTexture->shprog) );

  /* Set texture 0 environment mode */
  {
    CHECKGL( glActiveTexture(GL_TEXTURE0) );
    CHECKGL( glBindTexture(GL_TEXTURE_2D, ctk_render_target_get_color_buffer_ogl_id(rt_src)) );
  }

  {
    CHECKGL( glViewport(0, 0, window_w, window_h) );
    CHECKGL( glScissor(0, 0, window_w, window_h) );
    CHECKGL( glDisable(GL_SCISSOR_TEST) );
  }

  loc = glGetUniformLocation(g_shTexture->shprog, "Tex0");
  CHECKGL_MSG( "glGetUniformLocation" );
  CHECKGL( glUniform1i(loc, 0) );
  loc = glGetUniformLocation(g_shTexture->shprog, "color");
  CHECKGL( glUniform4f(loc, 1.0f, 1.0f, 1.0f, 1.0f) );

  glDisable(GL_BLEND);

  posId = glGetAttribLocation(g_shTexture->shprog, "a_position");
  texCoordId = glGetAttribLocation(g_shTexture->shprog, "a_texcoord0");

  CHECKGL( glVertexAttribPointer(posId, 4, GL_FLOAT, 0, 0, vertex_positions) );
  CHECKGL( glVertexAttribPointer(texCoordId, 2, GL_FLOAT, 0, 0, texture_coords) );

  CHECKGL( glEnableVertexAttribArray(posId) );
  CHECKGL( glEnableVertexAttribArray(texCoordId) );

  CHECKGL( glDrawArrays(GL_TRIANGLE_FAN, 0, 4) );

  CHECKGL( glDisableVertexAttribArray(posId) );
  CHECKGL( glDisableVertexAttribArray(texCoordId) );

  glEnable(GL_BLEND);

  /* Stop using the shader program */
  CHECKGL( glUseProgram(0) );

  /* Unbind texture */
  CHECKGL( glActiveTexture(GL_TEXTURE0) );
  CHECKGL( glBindTexture(GL_TEXTURE_2D, 0) );
}

void ctk_copy_render_target_to_cached_texture(CtkEffectContext *fxctx, CtkRenderTarget* rt_src, guint texid)
{
  CtkRenderTarget *rt_cache = ctk_effect_context_get_utility_render_target(fxctx);
  guint dest_fbo;

  guint rt_width = ctk_render_target_get_width(rt_src);
  guint rt_height = ctk_render_target_get_height(rt_src);

  /* make sure the target has valid color buffer and depth buffer */
  ctk_render_target_resize(rt_cache, rt_width, rt_height);

  dest_fbo = ctk_render_target_get_frame_buffer_ogl_id (rt_cache);

  CHECKGL (glBindFramebuffer (GL_FRAMEBUFFER, dest_fbo));

  CHECKGL (glActiveTexture(GL_TEXTURE0) );
  CHECKGL (glBindTexture (GL_TEXTURE_2D, texid) );
  CHECKGL (glTexImage2D (GL_TEXTURE_2D,
                             0,
                             GL_RGBA,
                             rt_width,
                             rt_height,
                             0,
                             GL_RGBA,
                             GL_UNSIGNED_BYTE,
                             NULL));
      
  CHECKGL (glFramebufferTexture2D (GL_FRAMEBUFFER,
                                      GL_COLOR_ATTACHMENT0,
                                      GL_TEXTURE_2D,
                                      texid,
                                      0));
 
  ctk_copy_rendertarget_to_rendertarget(rt_src, 0.0f, 0.0f, 1.0f, 1.0f, rt_cache, 0, 0, rt_width, rt_height); 
}

void
ctk_render_masked2_texture (guint             tex_id,
                            guint             tex_mask_id,
                            guint             texwidth,
                            guint             texheight,
                            ShaderProgram*    shader,
                            gint              window_w,
                            gint              window_h,
                            gint              x,
                            gint              y,
                            gint              w,
                            gint              h,
                            gfloat            Opacity)
{
  GLint loc;
  GLint posId, texCoord0Id, texCoord1Id;

  GLfloat texture_coords0[] =
  {
    0.0f, 1.0f,
    1.0f, 1.0f,
    1.0f, 0.0f,
    0.0f, 0.0f,
  };

  GLfloat texture_coords1[] =
  {
    0.0f, 1.0f,
    1.0f, 1.0f,
    1.0f, 0.0f,
    0.0f, 0.0f,
  };

  GLfloat vertex_positions[] =
  {
    (2.0*x-window_w)/window_w, -(2.0*y-window_h)/window_h, 0.0f, 1.0f,
    (2.0*(x+w)-window_w)/window_w, -(2.0*y-window_h)/window_h, 0.0f, 1.0f,
    (2.0*(x+w)-window_w)/window_w, -(2.0*(y+h)-window_h)/window_h, 0.0f, 1.0f,
    (2.0*x-window_w)/window_w, -(2.0*(y+h)-window_h)/window_h, 0.0f, 1.0f,
  };

  /* Set texture 0 environment mode */
  {
    CHECKGL( glActiveTexture(GL_TEXTURE0) );
    CHECKGL( glBindTexture(GL_TEXTURE_2D, tex_id) );
  }

  /*  Set texture 1 environment mode */
  {
    CHECKGL( glActiveTexture(GL_TEXTURE1) );
    CHECKGL( glBindTexture(GL_TEXTURE_2D, tex_mask_id) );
  }

  /* Use the shader */
  CHECKGL( glUseProgram(shader->shprog) );

  loc = glGetUniformLocation(g_shTexture->shprog, "Tex0");
  CHECKGL( glUniform1i(loc, 0) );
  loc = glGetUniformLocation(g_shTexture->shprog, "Tex1");
  CHECKGL( glUniform1i(loc, 1) );
  loc = glGetUniformLocation(shader->shprog, "color");
  CHECKGL( glUniform4f(loc, 1.0f, 1.0f, 1.0f, Opacity) );

  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  posId = glGetAttribLocation(shader->shprog, "a_position");
  texCoord0Id = glGetAttribLocation(shader->shprog, "a_texcoord0");
  texCoord1Id = glGetAttribLocation(shader->shprog, "a_texcoord1");

  CHECKGL( glVertexAttribPointer(posId, 4, GL_FLOAT, 0, 0, vertex_positions) );
  CHECKGL( glVertexAttribPointer(texCoord0Id, 2, GL_FLOAT, 0, 0, texture_coords0) );
  CHECKGL( glVertexAttribPointer(texCoord1Id, 2, GL_FLOAT, 0, 0, texture_coords1) );

  CHECKGL( glEnableVertexAttribArray(posId) );
  CHECKGL( glEnableVertexAttribArray(texCoord0Id) );
  CHECKGL( glEnableVertexAttribArray(texCoord1Id) );

  CHECKGL( glDrawArrays(GL_TRIANGLE_FAN, 0, 4) );

  CHECKGL( glDisableVertexAttribArray(posId) );
  CHECKGL( glDisableVertexAttribArray(texCoord0Id) );
  CHECKGL( glDisableVertexAttribArray(texCoord1Id) );

  /* Stop using the shader program */
  CHECKGL( glUseProgram(0) );

  /* Unbind texture */
  CHECKGL( glActiveTexture(GL_TEXTURE0) );
  CHECKGL( glBindTexture(GL_TEXTURE_2D, 0) );
  CHECKGL( glActiveTexture(GL_TEXTURE1) );
  CHECKGL( glBindTexture(GL_TEXTURE_2D, 0) );
}

void
compute_gaussian_kernel (gint iRadius,    /* radius of kernel                  */
                        gfloat fDeviation, /* deviation to use, 0.0f for auto   */
                        gfloat* W    /* pointer to array for coefficients */)
{
  guint num_tap = 7; /* must be odd number */

  gfloat sum = 0.0f;
  guint i = 0;

  float sigma = 1.0f;
  guint half = (num_tap-1)/2;

  W[half] = (1.0f/(sqrt(2.0f*3.14159265358f)*sigma)) * exp(0.0f/(2.0f*sigma*sigma));
  sum += W[half];
  for(i = 0; i < half; i++)
  {
    float X = (i + 1)*(i + 1);
    W[half - i - 1] = W[half + i + 1] = (1.0f/(sqrt(2.0f*3.14159265358f)*sigma)) * exp(-X/(2.0f*sigma*sigma));
    sum += 2.0f * W[half - i - 1];
  }

  /* normalization */
  for(i = 0; i < num_tap; i++)
  {
    W[i] = W[i] / sum;
  }
}

void
opengl_blur_pass (guint pSrcTexture,      /* texture to read from      */
                  guint src_width,
                  guint src_height,
                  GLfloat fDeviation,                 /* std. deviation to use     */
                  gboolean HorizontalDirection,     /* do which pass of the blur */
                  GLint iWindowWidth,               /* width of window/context   */
                  GLint iWindowHeight,            /* height of window/context  */
                  gboolean invertTexCoord)
{
  GLint loc;
  GLint posId, texCoordId;
  gfloat hdelta = 0;
  gfloat vdelta = 0;
  static GLfloat afKernel[12];

  static GLfloat aafVertices[4][4];

  GLfloat texture_coords0[] = {
       0.0f, 1.0f,
       0.0f, 0.0f,
       1.0f, 0.0f,
       1.0f, 1.0f};

  /* Use the shader */
  CHECKGL( glUseProgram(g_shMultipassBlur->shprog) );

  CHECKGL (glActiveTexture (GL_TEXTURE0));
  CHECKGL (glBindTexture (GL_TEXTURE_2D, pSrcTexture));
  CHECKGL (glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST));
  CHECKGL (glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST));
  CHECKGL (glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE));
  CHECKGL (glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE));

  compute_gaussian_kernel (5, fDeviation, afKernel);
  if (HorizontalDirection == FALSE)
  {
    /* Vertical blur */
    hdelta = 0.0f;
    vdelta = 1.0f / src_height;
  }
  else
  {
    /* Horizontal blur */
    hdelta = 1.0f / src_width;
    vdelta = 0.0f;
  }

  aafVertices[0][0] = -1.0f;
  aafVertices[0][1] = invertTexCoord?(2*src_height-iWindowHeight)/iWindowHeight:-1.0f;
  aafVertices[0][2] = 0.0f;
  aafVertices[0][3] = 1.0f;

  aafVertices[1][0] = -1.0f;
  aafVertices[1][1] = invertTexCoord?-1.0f:(2*src_height-iWindowHeight)/iWindowHeight;
  aafVertices[1][2] = 0.0f;
  aafVertices[1][3] = 1.0f;

  aafVertices[2][0] = (2*src_width-iWindowWidth)/iWindowWidth;
  aafVertices[2][1] = invertTexCoord?-1.0f:(2*src_height-iWindowHeight)/iWindowHeight;
  aafVertices[2][2] = 0.0f;
  aafVertices[2][3] = 1.0f;

  aafVertices[3][0] = (2*src_width-iWindowWidth)/iWindowWidth;
  aafVertices[3][1] = invertTexCoord?(2*src_height-iWindowHeight)/iWindowHeight:-1.0f;
  aafVertices[3][2] = 0.0f;
  aafVertices[3][3] = 1.0f;

  loc = glGetUniformLocation(g_shMultipassBlur->shprog, "Tex0");
  CHECKGL( glUniform1i(loc, 0) );
  loc = glGetUniformLocation(g_shMultipassBlur->shprog, "hdelta");
  CHECKGL( glUniform1f(loc, hdelta) );
  loc = glGetUniformLocation(g_shMultipassBlur->shprog, "vdelta");
  CHECKGL( glUniform1f(loc, vdelta) );
  loc = glGetUniformLocation(g_shMultipassBlur->shprog, "factors");
  CHECKGL( glUniform1fv(loc, 7, afKernel) );

  posId = glGetAttribLocation(g_shMultipassBlur->shprog, "a_position");
  CHECKGL( glVertexAttribPointer(posId, 4, GL_FLOAT, 0, 0, aafVertices) );
  CHECKGL( glEnableVertexAttribArray(posId) );

  texCoordId = glGetAttribLocation(g_shMultipassBlur->shprog, "a_texcoord0");
  CHECKGL( glVertexAttribPointer(texCoordId, 2, GL_FLOAT, 0, 0, texture_coords0) );
  CHECKGL( glEnableVertexAttribArray(texCoordId) );

  CHECKGL( glDrawArrays(GL_TRIANGLE_FAN, 0, 4) );

  CHECKGL( glDisableVertexAttribArray(posId) );
  CHECKGL( glDisableVertexAttribArray(texCoordId) );

  /* Stop using the shader program */
  CHECKGL( glUseProgram(0) );

  /* Unbind texture */
  CHECKGL( glActiveTexture(GL_TEXTURE0) );
  CHECKGL( glBindTexture(GL_TEXTURE_2D, 0) );
}

void ctk_render_texture_part(guint texid, gint x, gint y, gint width, gint height, float u0, float v0, float u1, float v1, int window_w, int window_h)
{
  GLint loc;
  GLint posId, texCoordId;

  GLfloat texture_coords[] =
  {
    u0, v0,
    u0, v1,
    u1, v1,
    u1, v0,
  };

  GLfloat vertex_positions[] =
  {
    (2.0*x-window_w)/window_w, -(2.0*y-window_h)/window_h, 0.0f, 1.0f,
    (2.0*x-window_w)/window_w, -(2.0*(y+height)-window_h)/window_h, 0.0f, 1.0f,
    (2.0*(x+width)-window_w)/window_w, -(2.0*(y+height)-window_h)/window_h, 0.0f, 1.0f,
    (2.0*(x+width)-window_w)/window_w, -(2.0*y-window_h)/window_h, 0.0f, 1.0f,
  };

  /* Set texture 0 environment mode */
  {
    CHECKGL( glActiveTexture(GL_TEXTURE0) );
    CHECKGL( glBindTexture(GL_TEXTURE_2D, texid) );
  }

  /* Use the shader */
  CHECKGL( glUseProgram(g_shTexture->shprog) );

  /* Save the blending states */
  glEnable(GL_BLEND);

  loc = glGetUniformLocation(g_shTexture->shprog, "Tex0");
  CHECKGL( glUniform1i(loc, 0) );
  loc = glGetUniformLocation(g_shTexture->shprog, "color");
  CHECKGL( glUniform4f(loc, 1.0f, 1.0f, 1.0f, 1.0f) );

  posId = glGetAttribLocation(g_shTexture->shprog, "a_position");
  texCoordId = glGetAttribLocation(g_shTexture->shprog, "a_texcoord0");

  CHECKGL( glVertexAttribPointer(posId, 4, GL_FLOAT, 0, 0, vertex_positions) );
  CHECKGL( glVertexAttribPointer(texCoordId, 2, GL_FLOAT, 0, 0, texture_coords) );

  CHECKGL( glEnableVertexAttribArray(posId) );
  CHECKGL( glEnableVertexAttribArray(texCoordId) );

  CHECKGL( glDrawArrays(GL_TRIANGLE_FAN, 0, 4) );

  CHECKGL( glDisableVertexAttribArray(posId) );
  CHECKGL( glDisableVertexAttribArray(texCoordId) );

  /* Stop using the shader program */
  CHECKGL( glUseProgram(0) );
  CHECKGL( glActiveTexture (GL_TEXTURE0) );
  CHECKGL( glBindTexture (GL_TEXTURE_2D, 0) )
}

void
ctk_render_texture_part_with_mask (guint          tex_id,
                                guint             tex_mask_id,
                                gint              x,
                                gint              y,
                                gint              width,
                                gint              height,
                                float u0, float v0, float u1, float v1,
                                float U0, float V0, float U1, float V1,
                                gint              window_w,
                                gint              window_h,
                                gfloat            opacity)
{
  GLint loc;
  GLint posId, texCoordId;

  GLfloat texture_coords0[] =
  {
    u0, v0,
    u0, v1,
    u1, v1,
    u1, v0,
  };

  GLfloat texture_coords1[] =
  {
    U0, V0,
    U0, V1,
    U1, V1,
    U1, V0,
  };

  GLfloat vertex_positions[] =
  {
    (2.0*x-window_w)/window_w, -(2.0*y-window_h)/window_h, 0.0f, 1.0f,
    (2.0*x-window_w)/window_w, -(2.0*(y+height)-window_h)/window_h, 0.0f, 1.0f,
    (2.0*(x+width)-window_w)/window_w, -(2.0*(y+height)-window_h)/window_h, 0.0f, 1.0f,
    (2.0*(x+width)-window_w)/window_w, -(2.0*y-window_h)/window_h, 0.0f, 1.0f,
  };

  /* Set texture 0 environment mode */
  {
    CHECKGL( glActiveTexture(GL_TEXTURE0) );
    CHECKGL( glBindTexture(GL_TEXTURE_2D, tex_id) );
  }

    /*  Set texture 1 environment mode */
  {
    CHECKGL( glActiveTexture(GL_TEXTURE1) );
    CHECKGL( glBindTexture(GL_TEXTURE_2D, tex_mask_id) );
  }

  /* Use the shader */
  CHECKGL( glUseProgram(g_shTexture->shprog) );

  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  loc = glGetUniformLocation(g_shTexture->shprog, "Tex0");
  CHECKGL( glUniform1i(loc, 0) );
  loc = glGetUniformLocation(g_shTexture->shprog, "Tex1");
  CHECKGL( glUniform1i(loc, 1) );
  loc = glGetUniformLocation(g_shTexture->shprog, "color");
  CHECKGL( glUniform4f(loc, 1.0f, 1.0f, 1.0f, opacity) );

  posId = glGetAttribLocation(g_shTexture->shprog, "a_position");
  CHECKGL( glVertexAttribPointer(posId, 4, GL_FLOAT, 0, 0, vertex_positions) );
  CHECKGL( glEnableVertexAttribArray(posId) );

  texCoordId = glGetAttribLocation(g_shTexture->shprog, "a_texcoord0");
  CHECKGL( glVertexAttribPointer(texCoordId, 2, GL_FLOAT, 0, 0, texture_coords0) );
  CHECKGL( glEnableVertexAttribArray(texCoordId) );

  texCoordId = glGetAttribLocation(g_shTexture->shprog, "a_texcoord1");
  CHECKGL( glVertexAttribPointer(texCoordId, 2, GL_FLOAT, 0, 0, texture_coords1) );
  CHECKGL( glEnableVertexAttribArray(texCoordId) );

  CHECKGL( glDrawArrays(GL_TRIANGLE_FAN, 0, 4) );

  CHECKGL( glDisableVertexAttribArray(posId) );
  CHECKGL( glDisableVertexAttribArray(texCoordId) );

  /* Stop using the shader program */
  CHECKGL( glUseProgram(0) );

  /* Unbind texture */
  CHECKGL( glActiveTexture(GL_TEXTURE0) );
  CHECKGL( glBindTexture(GL_TEXTURE_2D, 0) );
  CHECKGL( glActiveTexture(GL_TEXTURE1) );
  CHECKGL( glBindTexture(GL_TEXTURE_2D, 0) );
}

#endif

void ctk_get_actor_screen_position (CtkActor* actor, gfloat *x, gfloat *y, gfloat *width, gfloat *height, ClutterVertex vertex[4])
{
  float viewport[4];
  float X, Y, Z, W;

  ClutterActorBox box;

  CoglMatrix mod_mtx;
  CoglMatrix proj_mtx;

  cogl_push_matrix();

  cogl_get_modelview_matrix (&mod_mtx);
  cogl_get_projection_matrix (&proj_mtx);
  cogl_get_viewport (viewport);

  cogl_pop_matrix();

  ctk_actor_recurse_get_stored_allocation_box ( actor, &box);

  *width =  box.x2 - box.x1;
  *height = box.y2 - box.y1;


  X = 0.0f; Y = 0.0f; Z = 0.0f; W = 1.0f;
  cogl_matrix_transform_point (&mod_mtx, &X, &Y, &Z, &W);
  cogl_matrix_transform_point (&proj_mtx, &X, &Y, &Z, &W);
  vertex[0].x = (((X/W) + 1.0f)/2)*viewport[2] + viewport[0];
  vertex[0].y = ((1.0f - (Y/W))/2)*viewport[3] + viewport[1];
  
  X = 0.0f; Y = *height; Z = 0.0f; W = 1.0f;
  cogl_matrix_transform_point (&mod_mtx, &X, &Y, &Z, &W);
  cogl_matrix_transform_point (&proj_mtx, &X, &Y, &Z, &W);
  vertex[1].x = (((X/W) + 1.0f)/2)*viewport[2] + viewport[0];
  vertex[1].y = ((1.0f - (Y/W))/2)*viewport[3] + viewport[1];

  X = *width; Y = *height; Z = 0.0f; W = 1.0f;
  cogl_matrix_transform_point (&mod_mtx, &X, &Y, &Z, &W);
  cogl_matrix_transform_point (&proj_mtx, &X, &Y, &Z, &W);
  vertex[2].x = (((X/W) + 1.0f)/2)*viewport[2] + viewport[0];
  vertex[2].y = ((1.0f - (Y/W))/2)*viewport[3] + viewport[1];

  X = *width; Y = 0.0f; Z = 0.0f; W = 1.0f;
  cogl_matrix_transform_point (&mod_mtx, &X, &Y, &Z, &W);
  cogl_matrix_transform_point (&proj_mtx, &X, &Y, &Z, &W);
  vertex[3].x = (((X/W) + 1.0f)/2)*viewport[2] + viewport[0];
  vertex[3].y = ((1.0f - (Y/W))/2)*viewport[3] + viewport[1];

  float x_min, x_max, y_min, y_max;
  x_min = x_max = vertex[0].x;
  y_min = y_max = vertex[0].y;
  int i = 0;
  for (i = 1; i < 4; i++)
    {
      if (x_min > vertex[i].x)
        x_min = vertex[i].x;
      if (x_max < vertex[i].x)
        x_max = vertex[i].x;

      if (y_min > vertex[i].y)
        y_min = vertex[i].y;
      if (y_max < vertex[i].y)
        y_max = vertex[i].y;
    }

 /* X, y represents the position of the boundung box surounding the actor on screen 
  * while, width and height are the dimensions of the bounding box
  */

  *x = x_min;
  *y = y_min;

  *width =  x_max - x_min;
  *height = y_max - y_min;
}


